from unittest import TestCase
import numpy as np
from dianna.methods.kernelshap import KERNELSHAPImage


class ShapOnImages(TestCase):
    """Suite of Kernelshap tests for the image case."""
    def test_shap_segment_image(self):
        """Test if the segmentation of images are correct given some data."""
        input_data = np.random.random((28, 28, 1))

        explainer = KERNELSHAPImage()
        # most arguments are chosen by default
        # https://scikit-image.org/docs/dev/api/skimage.segmentation.html#skimage.segmentation.slic
        n_segments = 50
        compactness = 10.0
        sigma = 0
        image_segments = explainer._segment_image(
            input_data,
            n_segments,
            compactness,
            sigma,
        )
        # check segments index
        assert np.amax(image_segments) <= n_segments
        # check image shape after segmentation
        assert image_segments.shape == input_data[:, :, 0].shape

    def test_shap_mask_image(self):
        """Test if the images masks are correct given some data."""
        input_data = np.random.random((28, 28, 1))
        explainer = KERNELSHAPImage()
        n_segments = 50
        compactness = 10.0
        sigma = 0
        background = 0
        segments_slic = explainer._segment_image(
            input_data,
            n_segments,
            compactness,
            sigma,
        )
        masked_image = explainer._mask_image(
            np.zeros((1, n_segments)), segments_slic, input_data, background,
        )
        # check if all points are masked
        assert np.array_equal(masked_image[0], np.zeros(input_data.shape))

    def test_shap_explain_image(self):
        """Tests exact expected output given an image and model for Kernelshap."""
        input_data = np.random.random((1, 28, 28))
        onnx_model_path = "./tests/test_data/mnist_model.onnx"
        n_segments = 50
        explainer = KERNELSHAPImage()
        labels = [0]
        shap_values, _ = explainer.explain(
            onnx_model_path,
            input_data,
            labels,
            nsamples=1000,
            background=0,
            n_segments=n_segments,
            compactness=10.0,
            sigma=0,
        )

        assert shap_values[0].shape == np.zeros((1, n_segments)).shape



import numpy as np
import pytest
import xarray as xr
from dianna import utils


def test_xarray_conversion_list():
    """Tests if xarray has correct dimensions and labels after creation."""
    data = np.zeros((1, 28, 28, 3))
    labels = ('batch', 'y', 'x', 'channels')

    data_x = utils.to_xarray(data, labels)

    assert data_x.shape == data.shape
    assert data_x.dims == labels


def test_xarray_conversion_dict():
    """Tests if xarray has correct dimensions and labels after creation with incomplete label dict."""
    data = np.zeros((1, 28, 28, 3))
    labels = {0: 'batch', -1: 'channels'}
    expected_labels = ('batch', 'dim_1', 'dim_2', 'channels')

    data_x = utils.to_xarray(data, labels)

    assert data_x.shape == data.shape
    assert data_x.dims == expected_labels


def test_xarray_conversion_required_label():
    """Tests if error is raised when omitting some required label."""
    data = np.zeros((1, 28, 28, 3))
    labels = ('batch', 'y', 'x', 'channels')
    required = ['non-existent']

    with pytest.raises(AssertionError):
        utils.to_xarray(data, labels, required_labels=required)


def test_xarray_move_axis():
    """Tests if moving an axis results in expected shape and dimensions."""
    data = xr.DataArray(np.zeros((4, 1, 28, 28)), dims=('batch', 'channels', 'y', 'x'))
    expected_labels = ('batch', 'y', 'x', 'channels')
    expected_shape = (4, 28, 28, 1)

    data_moved = utils.move_axis(data, 'channels', -1)

    assert data_moved.dims == expected_labels
    assert data_moved.shape == expected_shape


def test_xarray_move_axis_nonexistent():
    """Tests if error is raised when trying to move some nonexistent axis."""
    data = xr.DataArray(np.zeros((4, 1, 28, 28)), dims=('batch', 'channels', 'y', 'x'))
    nonexistent_label = 'foo'

    with pytest.raises(ValueError):
        utils.move_axis(data, nonexistent_label, 0)



from unittest import TestCase
import numpy as np
import dianna
import dianna.visualization
from dianna.methods.rise import RISEImage
from dianna.methods.rise import RISEText
from dianna.utils import get_function
from tests.utils import assert_explanation_satisfies_expectations
from tests.utils import get_mnist_1_data
from tests.utils import load_movie_review_model
from tests.utils import run_model
from .test_onnx_runner import generate_data


class RiseOnImages(TestCase):
    """Suite of RISE tests for the image case."""

    @staticmethod
    def test_rise_function():
        """Test if rise runs and outputs the correct shape given some data and a model function."""
        input_data = np.random.random((224, 224, 3))
        axis_labels = ['y', 'x', 'channels']
        labels = [1]
        heatmaps_expected = np.load('tests/test_data/heatmap_rise_function.npy')

        heatmaps = dianna.explain_image(run_model, input_data, "RISE", labels, axis_labels=axis_labels, n_masks=200,
                                        p_keep=.5)

        assert heatmaps[0].shape == input_data.shape[:2]
        assert np.allclose(heatmaps, heatmaps_expected, atol=1e-5)

    @staticmethod
    def test_rise_filename():
        """Test if rise runs and outputs the correct shape given some data and a model file."""
        model_filename = 'tests/test_data/mnist_model.onnx'
        input_data = generate_data(batch_size=1).astype(np.float32)[0]
        heatmaps_expected = np.load('tests/test_data/heatmap_rise_filename.npy')
        labels = [1]

        heatmaps = dianna.explain_image(model_filename, input_data, "RISE", labels, n_masks=200, p_keep=.5)

        assert heatmaps[0].shape == input_data.shape[1:]
        print(heatmaps_expected.shape)
        assert np.allclose(heatmaps, heatmaps_expected, atol=1e-5)

    @staticmethod
    def test_rise_determine_p_keep_for_images():
        """Tests exact expected p_keep given an image and model."""
        np.random.seed(0)
        expected_p_exact_keep = .4
        model_filename = 'tests/test_data/mnist_model.onnx'
        data = get_mnist_1_data().astype(np.float32)

        p_keep = RISEImage()._determine_p_keep(
            data, get_function(model_filename))

        assert np.isclose(p_keep, expected_p_exact_keep)


class RiseOnText(TestCase):
    """Suite of RISE tests for the text case."""

    def test_rise_text(self):
        """Tests exact expected output given a text and model."""
        review = 'such a bad movie'
        expected_words = ['such', 'a', 'bad', 'movie']
        expected_word_indices = [0, 1, 2, 3]
        expected_positive_scores = [0.30, 0.29, 0.04, 0.25]

        positive_explanation = dianna.explain_text(self.runner, review, tokenizer=self.runner.tokenizer,
                                                   labels=(1, 0), method='RISE', p_keep=.5)[0]

        assert_explanation_satisfies_expectations(positive_explanation, expected_positive_scores, expected_word_indices,
                                                  expected_words)

    def test_rise_determine_p_keep_for_text(self):
        """Tests exact expected p_keep given a text and model."""
        expected_p_exact_keep = .7
        input_text = 'such a bad movie'
        runner = get_function(self.runner)
        input_tokens = np.asarray(runner.tokenizer.tokenize(input_text))

        p_keep = RISEText()._determine_p_keep(input_tokens, runner,  # pylint: disable=protected-access
                                              runner.tokenizer)
        assert np.isclose(p_keep, expected_p_exact_keep)

    def setUp(self) -> None:
        """Set seed and load runner."""
        np.random.seed(0)
        self.runner = load_movie_review_model()



from unittest import TestCase
import numpy as np
import pytest
import dianna
import dianna.visualization
from dianna.methods.lime import LIMEImage
from tests.test_onnx_runner import generate_data
from tests.utils import assert_explanation_satisfies_expectations
from tests.utils import load_movie_review_model
from tests.utils import run_model


class LimeOnImages(TestCase):
    """Suite of Lime tests for the image case."""

    @staticmethod
    def test_lime_function():
        """Test if lime runs and outputs are correct given some data and a model function."""
        input_data = np.random.random((224, 224, 3))
        heatmap_expected = np.load('tests/test_data/heatmap_lime_function.npy')
        labels = [1]

        explainer = LIMEImage(random_state=42)
        heatmap = explainer.explain(run_model, input_data, labels, num_samples=100)

        assert heatmap[0].shape == input_data.shape[:2]
        assert np.allclose(heatmap, heatmap_expected, atol=1e-5)

    @staticmethod
    def test_lime_filename():
        """Test if lime runs and outputs are correct given some data and a model file."""
        model_filename = 'tests/test_data/mnist_model.onnx'
        input_data = generate_data(batch_size=1)[0].astype(np.float32)
        axis_labels = ('channels', 'y', 'x')
        labels = [1]

        heatmap = dianna.explain_image(model_filename, input_data, method="LIME", labels=labels, random_state=42,
                                       axis_labels=axis_labels)

        heatmap_expected = np.load('tests/test_data/heatmap_lime_filename.npy')
        assert heatmap[0].shape == input_data[0].shape
        assert np.allclose(heatmap, heatmap_expected, atol=1e-5)

    def setUp(self) -> None:
        """Set seed."""
        np.random.seed(42)


class LimeOnText(TestCase):
    """Suite of Lime tests for the text case."""

    def test_lime_text(self):
        """Tests exact expected output given a text and model for Lime."""
        review = 'such a bad movie'
        expected_words = ['bad', 'such', 'movie', 'a']
        expected_word_indices = [2, 0, 3, 1]
        expected_scores = [0.49226245, -0.04637814, 0.03648112, -0.00837716]

        explanation = dianna.explain_text(self.runner, review, tokenizer=self.runner.tokenizer,
                                          labels=[0], method='LIME', random_state=42)[0]

        assert_explanation_satisfies_expectations(explanation, expected_scores, expected_word_indices, expected_words)

    def test_lime_text_special_chars(self):
        """Tests exact expected output given a text with special characters and model for Lime."""
        review = 'such a bad movie "!?\'"'
        expected_words = ['bad', 'such', 'movie', 'a', '"!?\'"']
        expected_word_indices = [2, 0, 3, 1, 4]
        expected_scores = [0.49421639, -0.04616689, 0.04045723, -0.00912872, -0.00148593]

        explanation = dianna.explain_text(self.runner, review, tokenizer=self.runner.tokenizer,
                                          labels=[0], method='LIME', random_state=42)[0]

        assert_explanation_satisfies_expectations(explanation, expected_scores, expected_word_indices, expected_words)

    def setUp(self) -> None:
        """Load the movie review model."""
        self.runner = load_movie_review_model()


@pytest.mark.parametrize("review", [
    'review with !!!?',
    'review with! ?',
    'review with???!',
    'Review with Capital',
])
class TestLimeOnTextSpecialCharacters:
    """Regression tests for inputs with symbols for LIME (https://github.com/dianna-ai/dianna/issues/437)."""
    runner = load_movie_review_model()  # load model once for all tests in this class

    def test_lime_text_special_chars_regression_test(self, review):
        """Just don't raise an error on this input with special characters."""
        _ = dianna.explain_text(self.runner, review, tokenizer=self.runner.tokenizer,
                                labels=[0], method='LIME', random_state=0)



import numpy as np
from dianna.visualization import plot_timeseries


def test_plot_timeseries():
    """Test plot series."""
    segments = [
        {'index': 0, 'start': 0, 'stop': 2, 'weight':-0.6},
        {'index': 1, 'start': 2, 'stop': 4, 'weight':-0.3},
        {'index': 2, 'start': 4, 'stop': 6, 'weight': 0.0},
        {'index': 3, 'start': 6, 'stop': 8, 'weight': 0.4},
        {'index': 4, 'start': 8, 'stop': 10, 'weight': 0.7},
    ]

    x = np.linspace(0, 10, 20)
    y = np.sin(x)

    plot_timeseries(x=x, y=y, segments=segments, show_plot=False)



from typing import Any
from typing import Dict
from typing import List
from typing import Optional
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm


def plot_timeseries(
    x: np.ndarray,
    y: np.ndarray,
    segments: List[Dict[str, Any]],
    xlabel: str = 'x',
    ylabel: str = 'y',
    cmap: Optional[str] = None,
    show_plot: bool = False,
    output_filename: Optional[str] = None,
    ax: Optional[plt.Axes] = None,
) -> plt.Axes:
    """Plot timeseries with segments highlighted.

    Args:
        x (np.ndarray): X-values
        y (np.ndarray): Y-values
        segments (List[Dict[str, Any]]): Segment data, must be a list of
            dicts with the following keys: 'index', 'start', 'end',
            'weight'. Here, `index` is the index of the segment of feature,
            `start` and `end` determine the location of the
            segment, and `weight` determines the color.
        xlabel (str, optional): Label for the x-axis
        ylabel (str, optional): Label for the y-axis
        cmap (str, optional): Matplotlib colormap
        show_plot (bool, optional): Shows plot if true (for testing or writing
            plots to disk instead).
        output_filename (str, optional): Name of the file to save
            the plot to (optional).
        ax (plt.Axes, optional): Matplotlib axes object
    """
    assert len(x) == len(y)

    if not ax:
        _, ax = plt.subplots()

    ax.plot(x, y, label='Timeseries')
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    cmap = plt.get_cmap(cmap)

    norm = plt.Normalize(-1, 1)

    for segment in segments:
        start = segment['start']
        stop = segment['stop']
        weight = segment['weight']
        index = segment['index']

        color = cmap(norm(weight))

        ax.axvspan(start, stop, color=color, alpha=0.5)
        ax.text(start, max(y), str(index))

    plt.colorbar(cm.ScalarMappable(norm=norm, cmap=cmap),
                 ax=ax,
                 label='weights')

    if show_plot:
        plt.show()
    if output_filename:
        plt.savefig(output_filename)

    return ax



import numpy as np
from dianna.utils.onnx_runner import SimpleModelRunner


def generate_data(batch_size):
    """Generate a batch of random data."""
    return np.random.randint(0, 256, size=(batch_size, 1, 28, 28))  # MNIST shape


def test_onnx_runner():
    """Tests if the onnx runner can run a test model and make predictions."""
    filename = 'tests/test_data/mnist_model.onnx'
    n_classes = 2  # binary MNIST model
    batch_size = 3

    runner = SimpleModelRunner(filename)
    pred_onnx = runner(generate_data(batch_size).astype(np.float32))

    assert pred_onnx.shape == (batch_size, n_classes)



import warnings
from typing import Union
import numpy as np


def generate_masks(input_data: np.array, number_of_masks: int, p_keep: float = 0.5):
    """Generate masks for time series data given a probability of keeping any time step unmasked.

    Args:
        input_data: Timeseries data to be explained.
        number_of_masks: Number of masks to generate.
        p_keep: the probability that any value remains unmasked.

    Returns:
    Single array containing all masks where the first dimension represents the batch.
    """
    series_length = input_data.shape[0]
    number_of_steps_masked = _determine_number_of_steps_masked(p_keep, series_length)

    masked_data_shape = [number_of_masks] + list(input_data.shape)
    masks = np.ones(masked_data_shape, dtype=np.bool)
    for i in range(number_of_masks):
        steps_to_mask = np.random.choice(series_length, number_of_steps_masked, False)
        masks[i, steps_to_mask] = False
    return masks


def mask_data(data: np.array, masks: np.array, mask_type: Union[object, str]):
    """Mask data given using a set of masks.

    Args:
        data: Input data.
        masks: an array with shape [number_of_masks] + data.shape
        mask_type: Masking strategy.

    Returns:
    Single array containing all masked input where the first dimension represents the batch.
    """
    number_of_masks = masks.shape[0]
    input_data_batch = np.repeat(np.expand_dims(data, 0), number_of_masks, axis=0)
    result = np.empty(input_data_batch.shape)
    result[masks] = input_data_batch[masks]
    result[~masks] = _get_mask_value(data, mask_type)
    return result


def _get_mask_value(data: np.array, mask_type: str) -> int:
    """Calculates a masking value of the given type for the data."""
    if callable(mask_type):
        return mask_type(data)
    if mask_type == 'mean':
        return np.mean(data)
    raise ValueError(f'Unknown mask_type selected: {mask_type}')


def _determine_number_of_steps_masked(p_keep: float, series_length: int) -> int:
    user_requested_steps = int(np.round(series_length * (1 - p_keep)))
    if user_requested_steps == series_length:
        warnings.warn('Warning: p_keep chosen too low. Continuing with leaving 1 time step unmasked per mask.')
        return series_length - 1
    if user_requested_steps == 0:
        warnings.warn('Warning: p_keep chosen too high. Continuing with masking 1 time step per mask.')
        return 1
    return user_requested_steps



from IPython.display import HTML
from IPython.display import display


def highlight_text(explanation,
                   input_tokens,
                   show_plot=True,
                   output_html_filename=None,
                   max_opacity=.8):
    """Highlights a given text based on values in a given explanation object.

    Args:
        explanation: list of tuples of (word, index of word in original data, importance)
        input_tokens: list of all tokens (including those without importance)
        show_plot: Shows plot if true (for testing or writing plots to disk instead)
        output_html_filename: Name of the file to save the plot to (optional).
        max_opacity: Maximum opacity (0-1)

    Returns:
        None
    """
    output = _create_html(input_tokens, explanation, max_opacity)

    if output_html_filename:
        with open(output_html_filename, 'w', encoding='utf-8') as output_html_file:
            print(output, file=output_html_file)

    if show_plot:
        display(HTML(output))


def _create_html(input_tokens, explanation, max_opacity):
    max_importance = max(abs(item[2]) for item in explanation)
    explained_indices = [index for _, index, _ in explanation]
    highlighted_words = []
    for index, word in enumerate(input_tokens):
        # if word has an explanation, highlight based on that, otherwise
        # make it grey
        try:
            explained_index = explained_indices.index(index)
            importance = explanation[explained_index][2]
            highlighted_words.append(
                _highlight_word(word, importance, max_importance, max_opacity)
                )
        except ValueError:
            highlighted_words.append(f'<span style="background:rgba(128, 128, 128, 0.3)">{word}</span>')

    return '<html><body>' + ' '.join(highlighted_words) + '</body></html>'


def _highlight_word(word, importance, max_importance, max_opacity):
    opacity = max_opacity * abs(importance) / max_importance
    if importance > 0:
        color = f'rgba(255, 0, 0, {opacity:.2f})'
    else:
        color = f'rgba(0, 0, 255, {opacity:2f})'
    highlighted_word = f'<span style="background:{color}">{word}</span>'
    return highlighted_word



import numpy as np
import dianna
import dianna.visualization
from tests.utils import run_model


def test_common_RISE_image_pipeline():  # noqa: N802 ignore case
    """No errors thrown while creating a relevance map and visualizing it."""
    input_image = np.random.random((224, 224, 3))
    axis_labels = {-1: 'channels'}
    labels = [0, 1]

    heatmap = dianna.explain_image(run_model, input_image, "RISE", labels, axis_labels=axis_labels)[0]
    dianna.visualization.plot_image(heatmap, show_plot=False)
    dianna.visualization.plot_image(heatmap, original_data=input_image[0], show_plot=False)


def test_common_RISE_timeseries_pipeline():  # noqa: N802 ignore case
    """No errors thrown while creating a relevance map and visualizing it."""
    input_timeseries = np.random.random((31, 1))
    labels = [0]

    heatmap = dianna.explain_timeseries(run_model, input_timeseries, "RISE", labels)[0]
    heatmap_channel = heatmap[:, 0]
    segments = []
    for i in range(len(heatmap_channel) - 1):
        segments.append({
            'index': i,
            'start': i,
            'stop': i + 1,
            'weight': heatmap_channel[i]})
    dianna.visualization.plot_timeseries(range(len(heatmap_channel)), input_timeseries[:, 0], segments, show_plot=False)



import numpy as np
from tqdm import tqdm
from dianna import utils
from dianna.utils.maskers import generate_masks
from dianna.utils.maskers import mask_data


def _make_predictions(input_data, runner, batch_size):
    """Process the input_data with the model runner in batches and return the predictions."""
    number_of_masks = input_data.shape[0]
    batch_predictions = []
    for i in tqdm(range(0, number_of_masks, batch_size), desc='Explaining'):
        batch_predictions.append(runner(input_data[i:i + batch_size]))
    return np.concatenate(batch_predictions)


# Duplicate code from rise.py:
def normalize(saliency, n_masks, p_keep):
    """Normalizes salience by number of masks and keep probability."""
    return saliency / n_masks / p_keep


class RISETimeseries:
    """RISE implementation for timeseries adapted from the image version of RISE."""

    def __init__(self, n_masks=1000, feature_res=8, p_keep=0.5,
                 preprocess_function=None):
        """RISE initializer.

        Args:
            n_masks (int): Number of masks to generate.
            feature_res (int): Resolution of features in masks.
            p_keep (float): Fraction of input data to keep in each mask (Default: auto-tune this value).
            preprocess_function (callable, optional): Function to preprocess input data with
        """
        self.n_masks = n_masks
        self.feature_res = feature_res
        self.p_keep = p_keep
        self.preprocess_function = preprocess_function
        self.masks = None
        self.predictions = None

    def explain(self, model_or_function, input_timeseries, labels, batch_size=100, mask_type='mean'):
        """Runs the RISE explainer on images.

           The model will be called with masked timeseries,
           with a shape defined by `batch_size` and the shape of `input_data`.

        Args:
            model_or_function (callable or str): The function that runs the model to be explained _or_
                                                 the path to a ONNX model on disk.
            input_timeseries (np.ndarray): Input timeseries data to be explained
            batch_size (int): Batch size to use for running the model.
            labels (Iterable(int)): Labels to be explained
            mask_type: Masking strategy for masked values. Choose from 'mean' or a callable(input_timeseries)

        Returns:
            Explanation heatmap for each class (np.ndarray).
        """
        runner = utils.get_function(model_or_function, preprocess_function=self.preprocess_function)
        self.masks = generate_masks(input_timeseries, number_of_masks=self.n_masks, p_keep=self.p_keep)
        masked = mask_data(input_timeseries, self.masks, mask_type=mask_type)

        self.predictions = _make_predictions(masked, runner, batch_size)
        n_labels = self.predictions.shape[1]

        saliency = self.predictions.T.dot(self.masks.reshape(self.n_masks, -1)).reshape(n_labels,
                                                                                        *input_timeseries.shape)
        selected_saliency = saliency[labels]
        return normalize(selected_saliency, self.n_masks, self.p_keep)



import numpy as np
from skimage.transform import resize
from tqdm import tqdm
from dianna import utils

# To Do: remove this import when the method for different input type is splitted
from dianna.methods.rise_timeseries import RISETimeseries  # noqa: F401 ignore unused import


def normalize(saliency, n_masks, p_keep):
    """Normalizes salience by number of masks and keep probability."""
    return saliency / n_masks / p_keep


def _upscale(grid_i, up_size):
    return resize(grid_i, up_size, order=1, mode='reflect', anti_aliasing=False)


def _predict_in_batches(masked, runner):
    batch_size = 50
    predictions = []
    for i in range(0, len(masked), batch_size):
        current_input = masked[i:min(i + batch_size, len(masked))]
        current_predictions = runner(current_input)
        predictions.append(current_predictions)
    predictions = np.concatenate(predictions)
    return predictions


class RISEText:
    """RISE implementation for text based on https://github.com/eclique/RISE/blob/master/Easy_start.ipynb."""

    def __init__(self, n_masks=1000, feature_res=8, p_keep=None,
                 preprocess_function=None):
        """RISE initializer.

        Args:
            n_masks (int): Number of masks to generate.
            feature_res (int): Resolution of features in masks.
            p_keep (float): Fraction of input data to keep in each mask (Default: auto-tune this value).
            preprocess_function (callable, optional): Function to preprocess input data with
        """
        self.n_masks = n_masks
        self.feature_res = feature_res
        self.p_keep = p_keep
        self.preprocess_function = preprocess_function
        self.masks = None
        self.predictions = None

    def explain(self, model_or_function, input_text, labels, tokenizer=None,
                batch_size=100):
        """Runs the RISE explainer on text.

           The model will be called with masked versions of the input text.

        Args:
            model_or_function (callable or str): The function that runs the model to be explained _or_
                                                 the path to a ONNX model on disk.
            input_text (np.ndarray): Text to be explained
            tokenizer: Tokenizer class with tokenize and convert_tokens_to_string methods, and mask_token attribute
            labels (Iterable(int)): Labels to be explained
            batch_size (int): Batch size to use for running the model.

        Returns:
            Explanation heatmap for each class (np.ndarray).
        """
        if tokenizer is None:
            raise ValueError('Please provide a tokenizer to explain_text.')

        runner = utils.get_function(model_or_function, preprocess_function=self.preprocess_function)
        input_tokens = np.asarray(tokenizer.tokenize(input_text))
        num_tokens = len(input_tokens)
        active_p_keep = self._determine_p_keep(input_tokens, runner, tokenizer) if self.p_keep is None else self.p_keep
        input_shape = (num_tokens,)
        self.masks = self._generate_masks(input_shape, active_p_keep,
                                          self.n_masks)  # Expose masks for to make user inspection possible
        masked_sentences = self._create_masked_sentences(input_tokens, self.masks, tokenizer)
        saliencies = self._get_saliencies(runner, masked_sentences, num_tokens, batch_size, active_p_keep)
        return self._reshape_result(input_tokens, labels, saliencies)

    def _determine_p_keep(self, input_text, runner, tokenizer, n_masks=100):
        """See n_mask default value https://github.com/dianna-ai/dianna/issues/24#issuecomment-1000152233."""
        p_keeps = np.arange(0.1, 1.0, 0.1)
        stds = []
        for p_keep in p_keeps:
            std = self._calculate_mean_class_std(p_keep, runner, input_text, tokenizer, n_masks=n_masks)
            stds += [std]
        best_i = np.argmax(stds)
        best_p_keep = p_keeps[best_i]
        print(f'Rise parameter p_keep was automatically determined at {best_p_keep}')
        return best_p_keep

    def _calculate_mean_class_std(self, p_keep, runner, input_text, tokenizer,
                                  n_masks):
        masks = self._generate_masks(input_text.shape, p_keep, n_masks)
        masked = self._create_masked_sentences(input_text, masks, tokenizer)
        predictions = _predict_in_batches(masked, runner)
        std_per_class = predictions.std(axis=0)
        return np.max(std_per_class)

    def _generate_masks(self, input_shape, p_keep, n_masks):
        masks = np.random.choice(a=(True, False), size=(n_masks,) + input_shape, p=(p_keep, 1 - p_keep))
        return masks

    def _get_saliencies(self, runner, sentences, num_tokens, batch_size, p_keep):
        self.predictions = self._get_predictions(sentences, runner, batch_size)
        unnormalized_saliency = self.predictions.T.dot(self.masks.reshape(self.n_masks, -1)).reshape(-1, num_tokens)
        return normalize(unnormalized_saliency, self.n_masks, p_keep)

    @staticmethod
    def _reshape_result(input_tokens, labels, saliencies):
        word_indices = list(range(len(input_tokens)))
        return [list(zip(input_tokens, word_indices, saliencies[label])) for label in labels]

    def _get_predictions(self, sentences, runner, batch_size):
        predictions = []
        for i in tqdm(range(0, self.n_masks, batch_size), desc='Explaining'):
            predictions.append(runner(sentences[i:i + batch_size]))
        predictions = np.concatenate(predictions)
        return predictions

    def _create_masked_sentences(self, tokens, masks, tokenizer):
        tokens_masked_list = [
            [token if keep else tokenizer.mask_token for token, keep in zip(tokens, mask)]
            for mask in masks]
        masked_sentences = [tokenizer.convert_tokens_to_string(tokens_masked)
                            for tokens_masked in tokens_masked_list]
        return masked_sentences


class RISEImage:
    """RISE implementation for images based on https://github.com/eclique/RISE/blob/master/Easy_start.ipynb."""

    def __init__(self, n_masks=1000, feature_res=8, p_keep=None,
                 axis_labels=None, preprocess_function=None):
        """RISE initializer.

        Args:
            n_masks (int): Number of masks to generate.
            feature_res (int): Resolution of features in masks.
            p_keep (float): Fraction of input data to keep in each mask (Default: auto-tune this value).
            axis_labels (dict/list, optional): If a dict, key,value pairs of axis index, name.
                                               If a list, the name of each axis where the index
                                               in the list is the axis index
            preprocess_function (callable, optional): Function to preprocess input data with
        """
        self.n_masks = n_masks
        self.feature_res = feature_res
        self.p_keep = p_keep
        self.preprocess_function = preprocess_function
        self.masks = None
        self.predictions = None
        self.axis_labels = axis_labels if axis_labels is not None else []

    def explain(self, model_or_function, input_data, labels, batch_size=100):
        """Runs the RISE explainer on images.

           The model will be called with masked images,
           with a shape defined by `batch_size` and the shape of `input_data`.

        Args:
            model_or_function (callable or str): The function that runs the model to be explained _or_
                                                 the path to a ONNX model on disk.
            input_data (np.ndarray): Image to be explained
            batch_size (int): Batch size to use for running the model.
            labels (Iterable(int)): Labels to be explained

        Returns:
            Explanation heatmap for each class (np.ndarray).
        """
        input_data, runner = self._prepare_input_data_and_model(input_data, model_or_function)

        active_p_keep = self._determine_p_keep(input_data, runner) if self.p_keep is None else self.p_keep

        # data shape without batch axis and channel axis
        img_shape = input_data.shape[1:3]
        # Expose masks for to make user inspection possible
        self.masks = self._generate_masks(img_shape, active_p_keep, self.n_masks)

        # Make sure multiplication is being done for correct axes
        masked = input_data * self.masks

        batch_predictions = []
        for i in tqdm(range(0, self.n_masks, batch_size), desc='Explaining'):
            batch_predictions.append(runner(masked[i:i + batch_size]))
        self.predictions = np.concatenate(batch_predictions)

        saliency = self.predictions.T.dot(self.masks.reshape(self.n_masks, -1)).reshape(-1, *img_shape)
        result = normalize(saliency, self.n_masks, active_p_keep)
        if labels is not None:
            result = result[list(labels)]
        return result

    def _prepare_input_data_and_model(self, input_data, model_or_function):
        """Prepares the input data as an xarray with an added batch dimension and creates a preprocessing function."""
        self._set_axis_labels(input_data)
        input_data = utils.to_xarray(input_data, self.axis_labels)
        input_data = input_data.expand_dims('batch', 0)
        input_data, full_preprocess_function = self._prepare_image_data(input_data)
        runner = utils.get_function(model_or_function, preprocess_function=full_preprocess_function)
        return input_data, runner

    def _set_axis_labels(self, input_data):
        # automatically determine the location of the channels axis if no axis_labels were provided
        axis_label_names = self.axis_labels.values() if isinstance(self.axis_labels, dict) else self.axis_labels
        if not axis_label_names:
            channels_axis_index = utils.locate_channels_axis(input_data.shape)
            self.axis_labels = {channels_axis_index: 'channels'}
        elif 'channels' not in axis_label_names:
            raise ValueError("When providing axis_labels it is required to provide the location"
                             " of the channels axis")

    def _determine_p_keep(self, input_data, runner, n_masks=100):
        """See n_mask default value https://github.com/dianna-ai/dianna/issues/24#issuecomment-1000152233."""
        p_keeps = np.arange(0.1, 1.0, 0.1)
        stds = []
        for p_keep in p_keeps:
            std = self._calculate_max_class_std(p_keep, runner, input_data, n_masks=n_masks)
            stds += [std]
        best_i = np.argmax(stds)
        best_p_keep = p_keeps[best_i]
        print(f'Rise parameter p_keep was automatically determined at {best_p_keep}')
        return best_p_keep

    def _calculate_max_class_std(self, p_keep, runner, input_data, n_masks):
        img_shape = input_data.shape[1:3]
        masks = self._generate_masks(img_shape, p_keep, n_masks)
        masked = input_data * masks
        predictions = _predict_in_batches(masked, runner)
        std_per_class = predictions.std(axis=0)
        return np.max(std_per_class)

    def _generate_masks(self, input_size, p_keep, n_masks):
        """Generates a set of random masks to mask the input data.

        Args:
            input_size (int): Size of a single sample of input data, for images without the channel axis.
            p_keep: ?
            n_masks: Number of masks

        Returns:
            The generated masks (np.ndarray)
        """
        cell_size = np.ceil(np.array(input_size) / self.feature_res)
        up_size = (self.feature_res + 1) * cell_size

        grid = np.random.choice(a=(True, False), size=(n_masks, self.feature_res, self.feature_res),
                                p=(p_keep, 1 - p_keep))
        grid = grid.astype('float32')

        masks = np.empty((n_masks, *input_size), dtype=np.float32)

        for i in range(n_masks):
            y = np.random.randint(0, cell_size[0])
            x = np.random.randint(0, cell_size[1])
            # Linear upsampling and cropping
            masks[i, :, :] = _upscale(grid[i], up_size)[y:y + input_size[0], x:x + input_size[1]]
        masks = masks.reshape(-1, *input_size, 1)
        return masks

    def _prepare_image_data(self, input_data):
        """Transforms the data to be of the shape and type RISE expects.

        Args:
            input_data (xarray): Data to be explained

        Returns:
            transformed input data, preprocessing function to use with utils.get_function()
        """
        # ensure channels axis is last and keep track of where it was so we can move it back
        channels_axis_index = input_data.dims.index('channels')
        input_data = utils.move_axis(input_data, 'channels', -1)
        # create preprocessing function that puts model input generated by RISE into the right shape and dtype,
        # followed by running the user's preprocessing function
        full_preprocess_function = self._get_full_preprocess_function(channels_axis_index, input_data.dtype)
        return input_data, full_preprocess_function

    def _get_full_preprocess_function(self, channel_axis_index, dtype):
        """Creates a full preprocessing function.

        Creates a preprocessing function that incorporates both the (optional) user's
        preprocessing function, as well as any needed dtype and shape conversions

        Args:
            channel_axis_index (int): Axis index of the channels in the input data
            dtype (type): Data type of the input data (e.g. np.float32)

        Returns:
            Function that first ensures the data has the same shape and type as the input data,
            then runs the users' preprocessing function
        """

        def moveaxis_function(data):
            return utils.move_axis(data, 'channels', channel_axis_index).astype(dtype).values

        if self.preprocess_function is None:
            return moveaxis_function
        return lambda data: self.preprocess_function(moveaxis_function(data))



# flake8: noqa: F401
"""DIANNA utilities."""
from .misc import get_function
from .misc import get_kwargs_applicable_to_function
from .misc import locate_channels_axis
from .misc import move_axis
from .misc import onnx_model_node_loader
from .misc import to_xarray
from .onnx_runner import SimpleModelRunner

__all__ = [
    'get_function',
    'get_kwargs_applicable_to_function',
    'locate_channels_axis',
    'move_axis',
    'onnx_model_node_loader',
    'to_xarray',
    'SimpleModelRunner',
]


import string
from abc import ABC
from abc import abstractmethod
from typing import List
import numpy as np


try:
    from torchtext.data import get_tokenizer
except ImportError as err:
    raise ImportError("Failed to import torchtext, please install manually or reinstall dianna with "
                      "text support: `pip install dianna[text]`") from err


class Tokenizer(ABC):
    """Abstract base class for tokenizing.

    Has the same interface as (part of) the transformers Tokenizer class.
    """
    def __init__(self, mask_token: str):
        """Tokenizer initializer.

        Args:
            mask_token (str): Token used as mask
        """
        self.mask_token = mask_token

    @abstractmethod
    def tokenize(self, sentence: str) -> List[str]:
        """Split sentence into list of tokens."""

    @abstractmethod
    def convert_tokens_to_string(self, tokens: List[str]) -> str:
        """Merge list of tokens back to sentence."""


class SpacyTokenizer(Tokenizer):
    """Spacy tokenizer for natural language."""
    def __init__(self, name: str = 'en_core_web_sm', mask_token: str = "UNKWORDZ"):
        """Spacy tokenizer initalizer.

        Args:
            name: Name of the Spacy tokenizer to use
            mask_token (str): Token used as mask
        """
        super().__init__(mask_token)
        self.spacy_tokenizer = get_tokenizer('spacy', name)

    def tokenize(self, sentence: str) -> List[str]:
        """Tokenize sentence."""
        raw_tokens = self.spacy_tokenizer(sentence)
        # do not consider several special characters in a row as separate tokens
        # special characters in string.punctuation are !"#$%&'()*+,-./:;<=>?@[\]^_`{|}~.
        # find indices of tokens that are a special character
        indices_in_tokens = np.where([token in string.punctuation for token in raw_tokens])[0]
        indices_in_raw_string = np.where([character in string.punctuation for character in (sentence)])[0]
        if indices_in_tokens.size == 0:
            # no special characters at all
            return raw_tokens

        # reconstruct list of tokens, combining consecutive special characters
        tokens = []
        steps_in_raw_string = np.diff(indices_in_raw_string)
        special_char_index = 0
        for idx, token in enumerate(raw_tokens):
            if idx not in indices_in_tokens:
                # not a special character
                tokens.append(token)
            elif special_char_index == 0 or steps_in_raw_string[special_char_index - 1] != 1:
                # create new token if this is the first special character,
                # or it does not directly follow another special character
                # in the original input
                tokens.append(token)
                special_char_index += 1
            else:
                # add token to existing one if both are special characters and
                # they are consecutive in the original input
                tokens[-1] = tokens[-1] + token
                special_char_index += 1
        return tokens

    def convert_tokens_to_string(self, tokens: List[str]) -> str:
        """Paste together with spaces in between."""
        sentence = " ".join(tokens)
        return sentence



import os
import shutil
import unittest
from pathlib import Path
import spacy
from dianna.utils.tokenizers import SpacyTokenizer
from dianna.visualization.text import highlight_text


spacy.cli.download('en_core_web_sm')


class TextExample:
    """Text and explanation for running visualizing tests."""
    original_text = 'Doloremque aliquam totam ut. Aspernatur repellendus autem quia deleniti. Natus accusamus ' \
                    'doloribus et in quam officiis veniam et. '
    tokenizer = SpacyTokenizer()
    tokens = tokenizer.tokenize(original_text)
    explanation = [
                    ('aliquam', 1, 0.6450221),
                    ('et', 19, 0.63976675),
                    ('in', 15, 0.6397511),
                    ('Aspernatur', 5, 0.6329795),
                    ('totam', 2, 0.6242337),
                    ('quia', 8, 0.61881626),
                    ('quam', 16, 0.6173148),
                    ('Natus', 11, 0.60527676),
                    ('ut', 3, 0.60203224)
                    ]


class TextExampleWithExpectedHtml:
    """Short text and explanation and its expected html output after visualizing."""
    expected_html = '<html><body><span style="background:rgba(255, 0, 0, 0.59)">Such</span> ' \
                    '<span style="background:rgba(255, 0, 0, 0.62)">a</span> ' \
                    '<span style="background:rgba(255, 0, 0, 0.80)">bad</span> ' \
                    '<span style="background:rgba(255, 0, 0, 0.63)">movie</span> ' \
                    '<span style="background:rgba(128, 128, 128, 0.3)">.</span></body></html>\n'

    original_text = 'Such a bad movie.'
    tokenizer = SpacyTokenizer()
    tokens = tokenizer.tokenize(original_text)
    explanation = [
                    ('bad', 2, 0.9959058),
                    ('movie', 3, 0.78263557),
                    ('a', 1, 0.7753202),
                    ('Such', 0, 0.73788315)
                    ]


class TextVisualizationTestCase(unittest.TestCase):
    """Suite of tests for visualizing text given text and explanation data."""
    temp_folder = 'temp_text_visualization_test'
    html_file_path = str(Path(temp_folder) / 'output.html')

    def test_text_visualization_html_output_exists(self):
        """Test if any output is generated at all."""
        highlight_text(TextExample.explanation, TextExample.tokens,
                       output_html_filename=self.html_file_path)

        assert Path(self.html_file_path).exists()

    def test_text_visualization_html_output_contains_text(self):
        """Test if all words in the input are present in the output html."""
        highlight_text(TextExample.explanation, TextExample.tokens,
                       output_html_filename=self.html_file_path)

        assert Path(self.html_file_path).exists()
        with open(self.html_file_path, encoding='utf-8') as result_file:
            result = result_file.read()
        for word in TextExample.tokens:
            assert word in result

    def test_text_visualization_html_output_is_correct(self):
        """Test if exact html output of visualization is correct."""
        highlight_text(TextExampleWithExpectedHtml.explanation, TextExampleWithExpectedHtml.tokens,
                       output_html_filename=self.html_file_path)

        assert Path(self.html_file_path).exists()

        with open(self.html_file_path, encoding='utf-8') as result_file:
            result = result_file.read()
        assert result == TextExampleWithExpectedHtml.expected_html

    def test_text_visualization_show_plot(self):
        """Test if it runs while showing the plot."""
        highlight_text(TextExample.explanation, TextExample.tokens,
                       show_plot=True)

    def setUp(self) -> None:
        """Setup."""
        os.mkdir(self.temp_folder)

    def tearDown(self) -> None:
        """Tear down."""
        shutil.rmtree(self.temp_folder, ignore_errors=True)
